/*
 * Copyright (c) 2011, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef GRIPPER_CLICK_PLUGIN_H_
#define GRIPPER_CLICK_PLUGIN_H_


#include "pr2_gripper_click/GripperClickPickupAction.h"
#include "pr2_gripper_click/GripperClickPlaceAction.h"


namespace rviz {
class VisualizationManager;
}

namespace Ogre {
class SceneNode;
}

namespace pr2_gripper_click {


// Interface class for a plugin that hooks into the
// plugin switcher
template<class ActionClass>
class GripperClickPlugin
{
public:

  typedef typename ActionClass::_action_result_type::_result_type ResultType;
  typedef typename ActionClass::_action_goal_type::_goal_type GoalType;

  GripperClickPlugin( std::string name ) :
      name_( name ),
      description_changed_(false) {}

  virtual ~GripperClickPlugin(){}

  std::string getName() { return name_; }

  // use this to change the help text that appears in the bottom part of the window
  void setDescription( const std::string &description ) {
    description_ = description;
    description_changed_ = true;
  }

  // the interface all plugins have to provide:

  virtual bool setGoal( const typename GoalType::ConstPtr &goal ) = 0;

  // do the actual initialization, instantiate ogre elements etc.
  virtual void init( rviz::VisualizationManager *vis_manager, Ogre::SceneManager* scene_manager, Ogre::SceneNode* camera_node ) = 0;

  virtual void update(float wall_dt, float ros_dt) = 0;

  // @return true if a result is available
  virtual bool hasValidResult() = 0;

  // make the orge content visible / invisible
  virtual void show( ) = 0;
  virtual void hide( ) = 0;

  virtual void onRenderWindowMouseEvents( wxMouseEvent& event, Ogre::Ray &mouse_ray ) = 0;

  //get the current gripper params for visualisation
  virtual void getGripperParams( bool &visible, Ogre::Vector3 &position, Ogre::Quaternion &orientation, float &angle )
  {
    visible = false;
  }

  virtual bool getResult(ResultType &result) = 0;

  std::string& getDescription() { return description_; }
  bool descriptionChanged() { bool c=description_changed_; description_changed_=false; return c; }

private:

  std::string name_;

  std::string description_;
  bool description_changed_;
};

}

#endif /* GRIPPER_CLICK_PLUGIN_H_ */
