/*
 * Copyright (c) 2009, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef GRIPPER_CLICK_RVIZ_UI
#define GRIPPER_CLICK_RVIZ_UI

#include <string>

#include <sensor_msgs/Image.h>
#include <sensor_msgs/CameraInfo.h>
#include <stereo_msgs/DisparityImage.h>

#include <OGRE/OgreSceneNode.h>
#include <OGRE/OgreSceneManager.h>

#include <wx/event.h>

#include <rviz/visualization_manager.h>
#include "rviz_interaction_tools/gripper.h"
#include "rviz_interaction_tools/disparity_renderer.h"
#include <rviz/status_level.h>

#include "pr2_gripper_click/gripper_click_plugin.h"

#include "gripper_click_ui.h"

#ifdef DEBUG_DISPLAY
#include "rviz_interaction_tools/line.h"
#endif

namespace rviz_interaction_tools {
class ImageOverlay;
}

namespace rviz {
class WindowManagerInterface;
class RenderPanel;
class VisualizationManager;
}

namespace Ogre {
class Light;
}

namespace pr2_gripper_click {

// Displays the camera + disparity images in a window
// allows to switch between 'plugins' that provide their own mouse interaction
// and visual elements for the 3D display
template<class ActionClass>
class GripperClickRvizUI : public GripperClickFrameBase {

public:

  typedef typename ActionClass::_action_result_type::_result_type ResultType;
  typedef typename ActionClass::_action_goal_type::_goal_type GoalType;

  enum ActionState{ IDLE, ACCEPTED, CANCELED };

  GripperClickRvizUI(rviz::VisualizationManager *visualization_manager);

  virtual ~GripperClickRvizUI();

  // @return state of the user interaction (@see ActionState)
  ActionState getState(){return action_state_;}

  // update frame buffer
  virtual void update(float wall_dt, float ros_dt);

  // add a new plugin to the list. Will take ownership.
  void addPlugin( GripperClickPlugin<ActionClass>* plugin );

  // Pass a new goal to all plugins
  bool setGoal( const typename GoalType::ConstPtr &goal );

  // transfer control & show selected plugin
  void setActivePlugin( unsigned index );

  // @brief get result for ActionClass
  // @return false if there is no valid result for the current plugin
  bool getResult( ResultType &result );

  // show / hide the window
  void show();
  void hide();

  // status text, describes if there is some error (e.g. a tf error)
  std::string getStatusText() const { return status_text_; }
  rviz::status_levels::StatusLevel getStatusLevel() const { return status_level_; }

  // set a new image / depth combination to be displayed
  void setImage( const sensor_msgs::Image &image,
      const stereo_msgs::DisparityImage &disparity_image,
      const sensor_msgs::CameraInfo &camera_info );

protected:

  // try to retrieve tf transform, set error status otherwise
  bool getTransform( std::string frame, Ogre::Vector3 &position, Ogre::Quaternion &orientation );

  // change text displaed at bottom of window
  void setBottomLabelText( const std::string &text );

  // change rviz status level & text
  void setStatus(rviz::status_levels::StatusLevel level, const std::string& text);

  // handle mouse movements, clicks etc.
  virtual void onRenderWindowMouseEvents( wxMouseEvent& event );

  // listeners for buttons
  virtual void acceptButtonClicked( wxCommandEvent& );
  virtual void cancelButtonClicked( wxCommandEvent& );

  //called when the user selects a different plugin
  virtual void onPluginChoice( wxCommandEvent& event );

  ActionState action_state_;

  rviz::RenderPanel* render_panel_;

  Ogre::SceneManager* scene_manager_;
  Ogre::SceneNode* camera_node_;

  rviz_interaction_tools::DisparityRenderer *disparity_renderer_;

  rviz_interaction_tools::Gripper* gripper_;

  rviz_interaction_tools::Gripper* main_win_gripper_;
  Ogre::SceneNode* main_win_camera_node_;
  rviz_interaction_tools::DisparityRenderer *main_win_disparity_renderer_;

  rviz::VisualizationManager *vis_manager_;

  std::string camera_frame_;

  Ogre::Light* gripper_light_;

  rviz_interaction_tools::ImageOverlay *image_overlay_;

  std::vector< GripperClickPlugin<ActionClass>* > plugins_;
  unsigned active_plugin_;

  rviz::status_levels::StatusLevel status_level_;
  std::string status_text_;

#ifdef DEBUG_DISPLAY
  rviz_interaction_tools::Line *line_;
#endif

};



}

#endif
