/*
 * Copyright (c) 2009, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "pr2_gripper_click/gripper_click_3d_gripper.h"


#include <tf/transform_datatypes.h>

#include <geometry_msgs/PoseStamped.h>

#include <rviz/render_panel.h>
#include <rviz/mesh_loader.h>

#include <OGRE/OgreEntity.h>
#include <OGRE/OgreManualObject.h>

#include <rviz_interaction_tools/unique_string_manager.h>
#include <rviz_interaction_tools/image_tools.h>

namespace pr2_gripper_click {

GripperClick3DGripper::GripperClick3DGripper() :
  GripperClickPlugin<GripperClickPickupAction>("3D Gripper"),
  gripper_angle_(0),
  gripper_roll_(0),
  valid_transform_(false)
{
}


GripperClick3DGripper::~GripperClick3DGripper()
{
}


void GripperClick3DGripper::init( rviz::VisualizationManager *vis_manager, Ogre::SceneManager* scene_manager, Ogre::SceneNode* camera_node )
{
  rviz_interaction_tools::UniqueStringManager usm;

  scene_node_ = camera_node->createChildSceneNode();

  line_object_ = scene_manager->createManualObject( usm.unique("GripperClickLineIndicator") );
  line_object_->setDynamic(true);
  line_object_->begin("BaseWhiteNoLighting", Ogre::RenderOperation::OT_LINE_STRIP);
  line_object_->position(0.0, 0.0, 0.0);
  line_object_->position(0.0, 0.0, 0.0);
  line_object_->index(0);
  line_object_->index(1);
  line_object_->index(0);
  line_object_->end();

  Ogre::AxisAlignedBox aabInf;
  aabInf.setInfinite();
  line_object_->setBoundingBox(aabInf);

  //click indicator will be rendered last
  line_object_->setRenderQueueGroup(Ogre::RENDER_QUEUE_OVERLAY);

  scene_node_->attachObject(line_object_);
}

bool GripperClick3DGripper::setGoal( const GripperClickPickupGoal::ConstPtr &goal )
{
  camera_info_ = goal->sensor_data.camera_info;
  disparity_image_ = goal->sensor_data.disparity_image;
  valid_transform_ = false;

  setDescription("Please double-click on the desired object to set the initial gripper position.");
  return true;
}


void GripperClick3DGripper::show( )
{
  scene_node_->setVisible(true);
}

void GripperClick3DGripper::hide( )
{
  scene_node_->setVisible(false);
}

void GripperClick3DGripper::update(float wall_dt, float ros_dt)
{
}

void GripperClick3DGripper::getGripperParams( bool &visible, Ogre::Vector3 &position, Ogre::Quaternion &orientation, float &angle )
{
  visible = valid_transform_;
  if ( visible )
  {
    btTransform trans = computeOrientation();

    //move back the transform so we get the wrist roll link
    btTransform back_trans(btQuaternion(0, 0, 0, 1), btVector3(-0.14, 0, 0));
    trans = trans * back_trans;
    btQuaternion rot = trans.getRotation();
    btVector3 trl = trans.getOrigin();

    position = Ogre::Vector3(trl.x(), trl.y(), trl.z());
    orientation = Ogre::Quaternion(rot.w(), rot.x(), rot.y(), rot.z());

    angle = gripper_angle_;
  }
}


btTransform GripperClick3DGripper::computeOrientation()
{
  btTransform transform;
  float px, py, pz;
  rviz_interaction_tools::getPoint(disparity_image_, camera_info_, dlb_.y_, dlb_.x_, px, py, pz);
  transform.setOrigin(btVector3(px, py, pz));
  
  float dy = lb_.y_;
  float dx = lb_.x_;
  float dz = std::min(sqrt(dx*dx + dy*dy), 200.0);
  dz = sqrt( 200 * 200 - dz * dz);
  btVector3 x_dir(dx, dy, dz);
  x_dir.normalize();
  
  btVector3 y_dir(1, 0, 0);
  if (fabs(x_dir.dot(y_dir)) > 0.99) y_dir = btVector3(0, 1, 0);
  
  btVector3 z_dir = x_dir.cross(y_dir);
  z_dir.normalize();
  y_dir = z_dir.cross(x_dir);
  
  btMatrix3x3 rot( x_dir.x(), x_dir.y(), x_dir.z(), 
                   y_dir.x(), y_dir.y(), y_dir.z(), 
                   z_dir.x(), z_dir.y(), z_dir.z() );
  rot = rot.transpose();
  
  transform = btTransform(rot, transform.getOrigin());
  
  btTransform roll_trans( btQuaternion(btVector3(1,0,0), gripper_roll_) );
  transform = transform * roll_trans;

  btTransform approach_trans(btQuaternion(0, 0, 0, 1), btVector3(gripper_approach_, 0, 0));
  transform = transform * approach_trans;

  line_object_->beginUpdate(0);
  line_object_->position(px,py,pz);
  line_object_->position(transform.getOrigin().x(),transform.getOrigin().y(),transform.getOrigin().z());
  line_object_->index(0);
  line_object_->index(1);
  line_object_->index(0);
  line_object_->end();

  return transform;
}

void GripperClick3DGripper::onRenderWindowMouseEvents( wxMouseEvent& event, Ogre::Ray &mouse_ray )
{
  int x = event.GetX();
  int y = event.GetY();

  if (event.GetWheelRotation() >= event.GetWheelDelta())
  {
    gripper_roll_ += 0.2;
    if (gripper_roll_ > M_PI) gripper_roll_ -= M_PI;
  }

  if (event.GetWheelRotation() <= -event.GetWheelDelta())
  {
    gripper_roll_ -= 0.2;
    if (gripper_roll_ < -M_PI) gripper_roll_ += M_PI;
  }

  if (event.ButtonDClick(wxMOUSE_BTN_LEFT))
  {
    if (rviz_interaction_tools::hasDisparityValue(disparity_image_, y, x))
    {
      valid_transform_ = true;
      dlb_.x_ = x;
      dlb_.y_ = y;
      gripper_approach_ = 0;
      setDescription("Left-click and drag the mouse to change the approaching angle.\nRight-click and drag to move the gripper forward and backwards\nTurn the mouse wheel to change the gripper rotation.\nPress and turn the mouse wheel to adjust the gripper opening.\nYou can reset the position by double-clicking on the desired location.\n\nPress OK When you are finished.");
    }
  }

  if (event.ButtonDown(wxMOUSE_BTN_LEFT))
  {
    lb_.down_ = true;
    lbdown_.x_ = lb_.x_ + x; //store the intiial position for incremental changes
    lbdown_.y_ = lb_.y_ + y;
  }
  if (event.ButtonUp(wxMOUSE_BTN_LEFT)) lb_.down_ = false;

  if (event.ButtonDown(wxMOUSE_BTN_RIGHT))
  {
    rb_.down_ = true;
    rb_.x_ = x;
    rb_.y_ = y;
    rb_.ref_ = gripper_approach_;
  }
  if (event.ButtonUp(wxMOUSE_BTN_RIGHT)) rb_.down_ = false;

  if (event.ButtonDown(wxMOUSE_BTN_MIDDLE)) 
  {
    mb_.down_ = true;
    mb_.x_ = x;
    mb_.y_ = y;
    mb_.ref_ = gripper_angle_;
  }
  if (event.ButtonUp(wxMOUSE_BTN_MIDDLE)) mb_.down_ = false;

  if (event.Dragging())
  {
    if (rb_.down_) // change gripper approach distance
    {
      int dy = rb_.y_ - y;
      gripper_approach_ = rb_.ref_ + ((float)dy / 100.0) * 0.1;
    }
    if (mb_.down_) // change gripper angle
    {
      int dy = y - mb_.y_;
      float da = ((float)dy/100.0) * 0.541;
      gripper_angle_ = mb_.ref_ + da;
      gripper_angle_ = std::min(gripper_angle_ , (float)0.541);
      gripper_angle_ = std::max(gripper_angle_ , (float)0.0);
    }
    if (lb_.down_)  // change gripper orientation
    {
      //absolute changes
      //lb_.x_ = dlb_.x_ - x;
      //lb_.y_ = dlb_.y_ - y;

      //incremental changes - more intuitive
      lb_.x_ = lbdown_.x_ - x;
      lb_.y_ = lbdown_.y_ - y;
    }
  }

}

bool GripperClick3DGripper::getResult(GripperClickPickupResult &result)
{
  //do not do shift correction, as we want the transform for the robot
  btTransform transform = computeOrientation();
  //move back the transform so we get the wrist roll link
  btTransform back_trans(btQuaternion(0, 0, 0, 1), btVector3(-0.14, 0, 0));
  transform = transform * back_trans;
  tf::poseTFToMsg(transform, result.grasp_pose.pose);
  result.grasp_pose.header.stamp = ros::Time::now();
  result.grasp_pose.header.frame_id = disparity_image_.header.frame_id;
  result.gripper_opening = gripper_angle_ / 0.5 * 0.0857;
  ROS_INFO("gripper_angle_: %.3f, using gripper opening of %.3f", gripper_angle_, result.gripper_opening);
  return true;
}

}
