/*
 * Copyright (c) 2009, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "pr2_gripper_click/gripper_click_display.h"
#include "pr2_gripper_click/gripper_click_rviz_ui.h"

#include <rviz/window_manager_interface.h>
#include <rviz/render_panel.h>
#include <rviz/visualization_manager.h>

namespace pr2_gripper_click {

template<class ActionClass>
GripperClickDisplay<ActionClass>::GripperClickDisplay( std::string action_topic, const std::string& name, rviz::VisualizationManager* manager ) :
  Display( name, manager ),
  gripper_click_rviz_ui_(0),
  action_server_(0),
  action_topic_(action_topic)
{
}

template<class ActionClass>
GripperClickDisplay<ActionClass>::~GripperClickDisplay()
{
  delete gripper_click_rviz_ui_;
  delete action_server_;
}

template<class ActionClass>
void GripperClickDisplay<ActionClass>::onEnable()
{
  if ( !gripper_click_rviz_ui_ )
  {
    gripper_click_rviz_ui_ = new GripperClickRvizUI<ActionClass>(vis_manager_);
  }
  startActionServer( );
}

template<class ActionClass>
void GripperClickDisplay<ActionClass>::onDisable()
{
  stopActionServer();
}

template<class ActionClass>
void GripperClickDisplay<ActionClass>::update(float wall_dt, float ros_dt)
{
  if ( !action_server_->isActive() )
  {
    return;
  }

  gripper_click_rviz_ui_->update(wall_dt, ros_dt);

  //check if the user has clicked on ok/cancel.
  switch( gripper_click_rviz_ui_->getState() )
  {
    case GripperClickRvizUI<ActionClass>::IDLE:
      break;
    case GripperClickRvizUI<ActionClass>::CANCELED:
      abortAction();
      break;
    case GripperClickRvizUI<ActionClass>::ACCEPTED:
    {
      ResultType result;
      if ( !gripper_click_rviz_ui_->getResult( result ) )
      {
        ROS_ERROR( "Could not retrieve result!" );
        abortAction();
      }
      else
      {
        ROS_INFO( "Action was successful. Sending back result." );
        action_server_->setSucceeded( result );
        gripper_click_rviz_ui_->hide();
      }
      break;
    }
  }
}

template<class ActionClass>
void GripperClickDisplay<ActionClass>::startActionServer( )
{
  if ( action_server_ )
  {
    ROS_ERROR( "Action server already started!" );
    return;
  }

  ROS_INFO("Starting action server.");

  //create non-threaded action server
  action_server_ =
      new actionlib::SimpleActionServer<ActionClass>( update_nh_, action_topic_, false );

  action_server_->registerGoalCallback( boost::bind(&GripperClickDisplay::acceptNewGoal, this) );
  action_server_->registerPreemptCallback( boost::bind(&GripperClickDisplay::preemptAction, this) );

  action_server_->start();
}


template<class ActionClass>
void GripperClickDisplay<ActionClass>::stopActionServer()
{
  if ( !action_server_ )
  {
    ROS_ERROR("Action server cannot be stopped because it is not running.");
    return;
  }

  //if we're currently being active, we have to cancel everything, clean up & hide the window
  if ( action_server_->isActive() )
  {
    ROS_WARN("Aborting action goal.");
    action_server_->setAborted();
    //cleanupAndHide();
  }

  ROS_INFO("Stopping action server.");
  delete action_server_;
  action_server_ = 0;
}

template<class ActionClass>
void GripperClickDisplay<ActionClass>::acceptNewGoal()
{
  const typename GoalType::ConstPtr &goal = action_server_->acceptNewGoal();

  gripper_click_rviz_ui_->setImage( goal->sensor_data.image, goal->sensor_data.disparity_image, goal->sensor_data.camera_info );

  if ( !gripper_click_rviz_ui_->setGoal( goal ) )
  {
    ROS_ERROR( "Couldn't set new goal!" );
    action_server_->setAborted();
    gripper_click_rviz_ui_->hide();
  }
}

template<class ActionClass>
void GripperClickDisplay<ActionClass>::preemptAction()
{
  action_server_->setAborted();
  gripper_click_rviz_ui_->hide();
}

template<class ActionClass>
void GripperClickDisplay<ActionClass>::abortAction()
{
  ROS_WARN( "Action was canceled." );
  action_server_->setPreempted();
  gripper_click_rviz_ui_->hide();
}

//explicitly instantiate pickup/place classes
template class GripperClickDisplay<GripperClickPickupAction>;
template class GripperClickDisplay<GripperClickPlaceAction>;

}
