/*
 * Copyright (c) 2009, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "pr2_gripper_click/gripper_click_rviz_ui.h"

#include <rviz/window_manager_interface.h>
#include <rviz/render_panel.h>
#include <rviz/visualization_manager.h>
#include <rviz/validate_floats.h>
#include <rviz/frame_manager.h>
#include <rviz/common.h>

#include <ogre_tools/point_cloud.h>

#include <OGRE/OgreRenderWindow.h>
#include <OGRE/OgreSceneNode.h>
#include <OGRE/OgreSceneManager.h>
#include <OGRE/OgreRoot.h>
#include <OGRE/OgreRectangle2D.h>
#include <OGRE/OgreMaterialManager.h>
#include <OGRE/OgreTextureManager.h>
#include <OGRE/OgreLight.h>

#include <sensor_msgs/image_encodings.h>

#include "rviz_interaction_tools/image_tools.h"
#include "rviz_interaction_tools/image_overlay.h"
#include "rviz_interaction_tools/camera_tools.h"
#include "rviz_interaction_tools/disparity_renderer.h"
#include "rviz_interaction_tools/unique_string_manager.h"

using namespace rviz_interaction_tools;

namespace pr2_gripper_click {

template<class ActionClass>
GripperClickRvizUI<ActionClass>::GripperClickRvizUI(rviz::VisualizationManager *visualization_manager) :
  GripperClickFrameBase(visualization_manager->getWindowManager()->getParentWindow()),
  action_state_(IDLE),
  vis_manager_(visualization_manager),
  active_plugin_(-1)
{
  UniqueStringManager usm;

  //construct basic ogre scene
  scene_manager_ = Ogre::Root::getSingleton().createSceneManager( Ogre::ST_GENERIC, usm.unique("GripperClick") );
  //scene_manager_->showBoundingBoxes(true);

  scene_manager_->setShadowColour( Ogre::ColourValue(0.8,0.8,0.8) );
  scene_manager_->setShadowTechnique(Ogre::SHADOWTYPE_STENCIL_MODULATIVE);

  camera_node_ = scene_manager_->getRootSceneNode()->createChildSceneNode();

  //create gripper in our own window
  gripper_ = new rviz_interaction_tools::Gripper( scene_manager_, camera_node_  );
  gripper_->setDepthCheckEnabled(true);
  gripper_->setRenderQueueGroup( Ogre::RENDER_QUEUE_MAIN );

  gripper_light_ = scene_manager_->createLight(usm.unique("gripper_light"));

  gripper_light_->setType( Ogre::Light::LT_DIRECTIONAL );
  gripper_light_->setDirection( 0,-1,0 );

  //create gripper in main window
  main_win_camera_node_ = visualization_manager->getSceneManager()->getRootSceneNode()->createChildSceneNode();
  main_win_gripper_ = new rviz_interaction_tools::Gripper( visualization_manager->getSceneManager(), main_win_camera_node_ );
  main_win_gripper_->setColour(1, 1, 1, 0.9);
  image_overlay_ = new rviz_interaction_tools::ImageOverlay( camera_node_, Ogre::RENDER_QUEUE_BACKGROUND );

  disparity_renderer_ = new rviz_interaction_tools::DisparityRenderer( camera_node_, Ogre::RENDER_QUEUE_MAIN );
  main_win_disparity_renderer_ = new rviz_interaction_tools::DisparityRenderer( main_win_camera_node_, Ogre::RENDER_QUEUE_MAIN );

  //put a render panel into the window
  render_panel_ = new rviz::RenderPanel(this, false);

  render_panel_->SetSize( panel_->GetRect() );

  // notify us when the user does something with the mouse
  render_panel_->Connect( wxEVT_MOTION,
                          wxMouseEventHandler( GripperClickRvizUI::onRenderWindowMouseEvents ), NULL, this );
  render_panel_->Connect( wxEVT_LEFT_DOWN, 
                          wxMouseEventHandler( GripperClickRvizUI::onRenderWindowMouseEvents ), NULL, this );
  render_panel_->Connect( wxEVT_MIDDLE_DOWN, 
                          wxMouseEventHandler( GripperClickRvizUI::onRenderWindowMouseEvents ), NULL, this );
  render_panel_->Connect( wxEVT_RIGHT_DOWN, 
                          wxMouseEventHandler( GripperClickRvizUI::onRenderWindowMouseEvents ), NULL, this );
  render_panel_->Connect( wxEVT_LEFT_UP, 
                          wxMouseEventHandler( GripperClickRvizUI::onRenderWindowMouseEvents ), NULL, this );
  render_panel_->Connect( wxEVT_MIDDLE_UP, 
                          wxMouseEventHandler( GripperClickRvizUI::onRenderWindowMouseEvents ), NULL, this );
  render_panel_->Connect( wxEVT_RIGHT_UP, 
                          wxMouseEventHandler( GripperClickRvizUI::onRenderWindowMouseEvents ), NULL, this );
  render_panel_->Connect( wxEVT_MOUSEWHEEL, 
                          wxMouseEventHandler( GripperClickRvizUI::onRenderWindowMouseEvents ), NULL, this );
  render_panel_->Connect( wxEVT_LEFT_DCLICK, 
                          wxMouseEventHandler( GripperClickRvizUI::onRenderWindowMouseEvents ), NULL, this );
  
  render_panel_->createRenderWindow();
  render_panel_->initialize(scene_manager_, visualization_manager);

  //fill camera matrix with dummy values
  render_panel_->getCamera()->setFOVy(Ogre::Radian(0.518));
  render_panel_->getCamera()->setAspectRatio(640.0 / 480.0);
  render_panel_->getCamera()->setNearClipDistance( 0.01f );

  render_panel_->setAutoRender(false);
  render_panel_->getViewport()->setOverlaysEnabled(false);
  render_panel_->getViewport()->setClearEveryFrame(true);
  render_panel_->getRenderWindow()->setAutoUpdated(false);
  render_panel_->getRenderWindow()->setActive(true);

  hide();

#ifdef DEBUG_DISPLAY
  line_ = new rviz_interaction_tools::Line( scene_manager_, visualization_manager->getSceneManager()->getRootSceneNode() );
#endif
}

template<class ActionClass>
GripperClickRvizUI<ActionClass>::~GripperClickRvizUI()
{
  for ( size_t i=0; i<plugins_.size(); i++ )
  {
    delete plugins_[i];
  }

  render_panel_->getRenderWindow()->setActive(false);
  delete gripper_;
  delete render_panel_;
  delete image_overlay_;
  delete disparity_renderer_;
  delete main_win_disparity_renderer_;
  Hide();

  vis_manager_->getSceneManager()->destroySceneNode(main_win_camera_node_);
  delete main_win_gripper_;

  Ogre::Root::getSingleton().destroySceneManager( scene_manager_ );
}

template<class ActionClass>
void GripperClickRvizUI<ActionClass>::update(float wall_dt, float ros_dt)
{
  if ( active_plugin_ < plugins_.size() )
  {
    bool visible;
    Ogre::Vector3 position;
    Ogre::Quaternion orientation;
    float angle;

    if ( plugins_[active_plugin_]->descriptionChanged() )
    {
      setBottomLabelText( plugins_[active_plugin_]->getDescription() );
    }

    plugins_[active_plugin_]->getGripperParams( visible, position, orientation, angle );

    gripper_->setVisible(visible);
    main_win_gripper_->setVisible(visible);

    if (visible)
    {
      //update our window
      gripper_->setPosition( position );
      gripper_->setOrientation( orientation );
      gripper_->setGripperAngle( angle );

      //update main window
      main_win_gripper_->setPosition( position );
      main_win_gripper_->setOrientation( orientation );
      main_win_gripper_->setGripperAngle( angle );
    }

    plugins_[active_plugin_]->update( wall_dt, ros_dt );
    accept_button_->Enable( plugins_[active_plugin_]->hasValidResult() );
  }
  else
  {
    ROS_ERROR_STREAM( "Invalid plugin index found! # of plugins: " << plugins_.size() );
    active_plugin_ = 0;
    accept_button_->Enable( false );
  }

  render_panel_->getRenderWindow()->update();
}

template<class ActionClass>
void GripperClickRvizUI<ActionClass>::show()
{
  Show();
  main_win_camera_node_->setVisible(true);
}

template<class ActionClass>
void GripperClickRvizUI<ActionClass>::hide()
{
  Hide();
  main_win_camera_node_->setVisible(false);
}

template<class ActionClass>
void GripperClickRvizUI<ActionClass>::setImage( const sensor_msgs::Image &image,
    const stereo_msgs::DisparityImage &disparity_image,
    const sensor_msgs::CameraInfo &camera_info )
{
  image_overlay_->setImage( image, disparity_image );
  image_overlay_->update();

  disparity_renderer_->setDisparityImage( disparity_image, camera_info, &image );
  disparity_renderer_->update();

  main_win_disparity_renderer_->setDisparityImage( disparity_image, camera_info, &image );
  main_win_disparity_renderer_->update();

  camera_frame_ = image.header.frame_id;

  rviz_interaction_tools::updateCamera( render_panel_->getCamera(), camera_info );

  Ogre::Vector3 cam_position;
  Ogre::Quaternion cam_orientation;

  if ( getTransform(camera_frame_, cam_position, cam_orientation) )
  {

    main_win_camera_node_->setPosition(cam_position);
    main_win_camera_node_->setOrientation(cam_orientation);

    camera_node_->setPosition(cam_position);
    camera_node_->setOrientation(cam_orientation);

    // convert vision (Z-forward) frame to ogre frame (Z-out)
    cam_orientation = cam_orientation * Ogre::Quaternion(Ogre::Degree(180), Ogre::Vector3::UNIT_X);

    render_panel_->getCamera()->setPosition(cam_position);
    render_panel_->getCamera()->setOrientation(cam_orientation);
  }
  else
  {
    ROS_ERROR("Cannot transform to %s", camera_frame_.c_str());
    main_win_gripper_->setVisible(false);
  }
}

template<class ActionClass>
bool GripperClickRvizUI<ActionClass>::getTransform( std::string frame, Ogre::Vector3 &position, Ogre::Quaternion &orientation )
{
  if (!vis_manager_->getFrameManager()->getTransform(frame, ros::Time(), position, orientation, false))
  {
    std::string error;
    if (vis_manager_->getFrameManager()->transformHasProblems(frame, ros::Time(), error))
    {
      setStatus(rviz::status_levels::Error, error);
    }
    else
    {
      std::stringstream ss;
      ss << "Could not transform from [" << frame << "] to Fixed Frame [" << vis_manager_->getFrameManager()->getFixedFrame() << "] for an unknown reason";
      setStatus(rviz::status_levels::Error, ss.str());
    }
    return false;
  }

  setStatus(rviz::status_levels::Ok, "Transform OK");
  return true;
}


template<class ActionClass>
void GripperClickRvizUI<ActionClass>::setStatus(rviz::status_levels::StatusLevel level, const std::string& text)
{
  status_level_ = level;
  status_text_ = text;
  bottom_label_->SetLabel( wxString::FromAscii( text.c_str() ) );

  if ( status_level_ == rviz::status_levels::Ok && active_plugin_ < plugins_.size() )
  {
    bottom_label_->SetLabel( wxString::FromAscii( plugins_[active_plugin_]->getDescription().c_str() ) );
  }
}

template<class ActionClass>
void GripperClickRvizUI<ActionClass>::setBottomLabelText(const std::string& text)
{
  if ( status_level_ == rviz::status_levels::Ok )
  {
    bottom_label_->SetLabel( wxString::FromAscii( text.c_str() ) );
  }
}


template<class ActionClass>
bool GripperClickRvizUI<ActionClass>::setGoal( const typename GoalType::ConstPtr &goal )
{
  for ( size_t i=0; i<plugins_.size(); i++ )
  {
    if ( !plugins_[i]->setGoal( goal ) )
    {
      ROS_ERROR( "Couldn't set new goal on '%s'!", plugins_[i]->getName().c_str() );
      return false;
    }
  }

  action_state_ = IDLE;
  show();
  return true;
}


template<class ActionClass>
void GripperClickRvizUI<ActionClass>::addPlugin( GripperClickPlugin<ActionClass>* plugin )
{
  plugin->init( vis_manager_, scene_manager_, camera_node_ );
  plugin->hide();
  plugin_choice_->Append( wxString::FromAscii( plugin->getName().c_str() ) );
  plugins_.push_back( plugin );

  //auto-select the first plugin to be added
  if ( plugins_.size()==1 )
  {
    plugin_choice_->Select(0);
    setActivePlugin(0);
  }
}

template<class ActionClass>
void GripperClickRvizUI<ActionClass>::setActivePlugin( unsigned index )
{
  if ( index >= plugins_.size() )
  {
    ROS_ERROR( "Non-existent plugin index selected!" );
  }
  if ( index != active_plugin_ )
  {
    if ( active_plugin_ < plugins_.size() )
    {
      plugins_[active_plugin_]->hide();
    }

    plugins_[index]->show();
    active_plugin_ = index;
    setBottomLabelText( plugins_[index]->getDescription() );
  }
}


template<class ActionClass>
bool GripperClickRvizUI<ActionClass>::getResult( ResultType &result )
{
  if ( active_plugin_ >= plugins_.size() || !plugins_[active_plugin_]->hasValidResult() )
  {
    return false;
  }

  plugins_[active_plugin_]->getResult( result );
  return true;
}


template<class ActionClass>
void GripperClickRvizUI<ActionClass>::acceptButtonClicked( wxCommandEvent& )
{
  action_state_ = ACCEPTED;
  Hide();
}

template<class ActionClass>
void GripperClickRvizUI<ActionClass>::cancelButtonClicked( wxCommandEvent& )
{
  action_state_ = CANCELED;
  Hide();
}

template<class ActionClass>
void GripperClickRvizUI<ActionClass>::onPluginChoice( wxCommandEvent& event )
{
  setActivePlugin( event.GetInt() );
}

template<class ActionClass>
void GripperClickRvizUI<ActionClass>::onRenderWindowMouseEvents( wxMouseEvent& event )
{
  if ( active_plugin_ < plugins_.size() )
  {
    int x = event.GetX();
    int y = event.GetY();

    //convert to pixel coordinates in the original image resolution
    wxSize size = render_panel_->GetSize();
    x = floor(x * image_overlay_->getWidth() / size.GetWidth() );
    y = floor(y * image_overlay_->getHeight() / size.GetHeight() );

    wxMouseEvent event_normalized = event;

    event_normalized.m_x = x;
    event_normalized.m_y = y;

    Ogre::Ray mouse_ray = render_panel_->getCamera()->getCameraToViewportRay(event.GetX()/float(size.GetWidth()), event.GetY()/float(size.GetHeight()));

#ifdef DEBUG_DISPLAY
    ROS_INFO_STREAM( "mouse " << event.GetX()/float(size.GetWidth()) << " " << event.GetY()/float(size.GetHeight())
        << "\n origin " << mouse_ray.getOrigin() << "\n direction " << mouse_ray.getDirection()
        << "\n camera pos " << render_panel_->getCamera()->getPosition() );

    line_->setLimits( mouse_ray.getOrigin(), mouse_ray.getOrigin() + mouse_ray.getDirection() );
#endif

    plugins_[active_plugin_]->onRenderWindowMouseEvents( event_normalized, mouse_ray );
  }
}

//explicitly instantiate pickup/place classes
template class GripperClickRvizUI<GripperClickPickupAction>;
template class GripperClickRvizUI<GripperClickPlaceAction>;


}
