/*
 * Copyright (c) 2009, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <ros/ros.h>

#include <actionlib/client/simple_action_client.h>
#include <std_msgs/Empty.h>

#include "pr2_gripper_click/GripperClickPickupAction.h"
#include "pr2_gripper_click/GripperClickPlaceAction.h"

class GraspPlanningGripperClick
{
private:
  ros::NodeHandle root_nh_;

  ros::NodeHandle priv_nh_;

  actionlib::SimpleActionClient<pr2_gripper_click::GripperClickPickupAction> *pickup_action_client_;
  actionlib::SimpleActionClient<pr2_gripper_click::GripperClickPlaceAction> *place_action_client_;

  std::string pickup_action_topic_;
  std::string place_action_topic_;

  ros::Subscriber sub_;

  void triggerCb(const std_msgs::EmptyConstPtr& msg)
  {
    while(!pickup_action_client_->waitForServer(ros::Duration(2.0)) && priv_nh_.ok())
    {
      ROS_INFO("Waiting for action client");
    }
    
    pr2_gripper_click::GripperClickPickupGoal click_goal;
    ROS_INFO("Sending goal");
    pickup_action_client_->sendGoal(click_goal);
    while (!pickup_action_client_->waitForResult(ros::Duration(0.5)) && priv_nh_.ok())
    {
    }
    if (!priv_nh_.ok()) 
    {
      pickup_action_client_->cancelGoal();
      return;
    }
    if (pickup_action_client_->getState() != actionlib::SimpleClientGoalState::SUCCEEDED)
    {
      ROS_INFO("The gripper click grasp action has not succeeded;");
    }
    else
    {    
      ROS_INFO("Gripper click grasp has succeeded;");
    }
  }

public:
  GraspPlanningGripperClick() : 
    root_nh_(""), priv_nh_("~")
  {
    
    priv_nh_.param<std::string>("pickup_action_topic", pickup_action_topic_, "/gripper_click_pickup_ui");    
    pickup_action_client_ = new actionlib::SimpleActionClient<pr2_gripper_click::GripperClickPickupAction>
      (pickup_action_topic_, true);
    
    priv_nh_.param<std::string>("place_action_topic", place_action_topic_, "/gripper_click_place_ui");    
    place_action_client_ = new actionlib::SimpleActionClient<pr2_gripper_click::GripperClickPlaceAction>
      (place_action_topic_, true);

    //sub_ = root_nh_.subscribe("trigger", 1, boost::bind(&GraspPlanningGripperClick::triggerCb, this, _1));
    sub_ = root_nh_.subscribe("trigger", 1, &GraspPlanningGripperClick::triggerCb, this);

    ROS_INFO("Grasp planning gripper click node started");
  }
  ~GraspPlanningGripperClick()
  {
    delete pickup_action_client_;
    delete place_action_client_;
  }
};

int main(int argc, char **argv)
{
  ros::init(argc, argv, "grasp_planning_gripper_click_node");
  GraspPlanningGripperClick node;
  ros::spin();
  return 0;
}
